"""user.py contains basic information about the steno3d user"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties


class User(properties.HasProperties):
    """Class representing a user instance"""
    _model_api_location = "user"

    email = properties.GettableProperty(
        doc='Email',
    )
    name = properties.GettableProperty(
        doc='Name',
    )
    url = properties.GettableProperty(
        doc='URL',
    )
    affiliation = properties.GettableProperty(
        doc='Affiliation',
    )
    location = properties.GettableProperty(
        doc='Location',
    )
    username = properties.GettableProperty(
        doc='Username',
    )

    devel_key = properties.GettableProperty(
        doc='Developer API Key',
    )

    file_size_limit = properties.GettableProperty(
        doc='Inidividual file limit',
    )
    project_size_limit = properties.GettableProperty(
        doc='Project size limit',
    )
    project_resource_limit = properties.GettableProperty(
        doc='Maximum resources in a project',
    )

    def login_with_json(self, login_json):
        self._backend['username'] = login_json['uid']
        self._backend['email'] = login_json['email']
        self._backend['name'] = login_json['name']
        self._backend['url'] = login_json['url']
        self._backend['affiliation'] = login_json['affiliation']
        self._backend['location'] = login_json['location']
        self._backend['file_size_limit'] = login_json.get(
            'file_size_limit', 25000000
        )
        self._backend['project_size_limit'] = login_json.get(
            'project_size_limit', 100000000
        )
        self._backend['project_resource_limit'] = login_json.get(
            'project_resource_limit', 100
        )

    def set_key(self, devel_key):
        self._backend['devel_key'] = devel_key

    def logout(self):
        self._backend.pop('username', None)
        self._backend.pop('email', None)
        self._backend.pop('name', None)
        self._backend.pop('url', None)
        self._backend.pop('affiliation', None)
        self._backend.pop('location', None)
        self._backend.pop('devel_key', None)
        self._backend.pop('file_size_limit', None)
        self._backend.pop('project_size_limit', None)
        self._backend.pop('project_resource_limit', None)

    @property
    def logged_in(self):
        return self.username is not None



"""vector.py contains the Vector composite resource for Steno3D"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties

from .base import CompositeResource
from .options import ColorOptions

from .point import Mesh0D, _PointBinder
from .props import array_serializer, array_download


class _VectorOptions(ColorOptions):
    pass


class Vector(CompositeResource):
    """Contains all the information about a vector field"""

    _resource_class = 'vector'

    mesh = properties.Instance(
        doc='Mesh',
        instance_class=Mesh0D,
        auto_create=True,
    )
    vectors = properties.Array(
        doc='Vector',
        shape=('*', 3),
        dtype=float,
        serializer=array_serializer,
        deserializer=array_download(('*', 3), (float,)),
    )
    data = properties.List(
        doc='Data',
        prop=_PointBinder,
        coerce=True,
        required=False,
    )
    opts = properties.Instance(
        doc='Options',
        instance_class=_VectorOptions,
        auto_create=True,
    )

    def _nbytes(self):
        return (self.mesh._nbytes() + self.vectors.astype('f4').nbytes +
                sum(d.data._nbytes() for d in self.data))

    @properties.validator
    def _validate_data(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location == 'N'
            valid_length = self.mesh.nN
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'point.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return True

    @properties.validator
    def _validate_vectors(self):
        """Check if vectors is built correctly"""
        valid_length = self.mesh.nN
        if len(self.vectors) != valid_length:
            raise ValueError(
                'vectors length {datalen} does not match '
                'mesh vertex length {meshlen}'.format(
                    datalen=len(proposal['value']),
                    meshlen=valid_length
                )
            )
        return True

    def _get_dirty_files(self, force=False):
        files = {}
        dirty = self._dirty_props
        if 'vectors' in dirty or force:
            files['vectors'] = \
                self._props['vectors'].serialize(self.vectors)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        vec = super(Vector, cls)._build_from_json(json, **kwargs)
        vec.vectors = cls._props['vectors'].deserialize(
            json['vectors'],
        )
        return vec


__all__ = ['Vector']



"""data.py contains resource data structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
import random
from six import string_types

from numpy import ndarray
import properties

from .base import BaseData
from .props import array_serializer, array_download


class DataArray(BaseData):
    """Data array with unique values at every point in the mesh"""
    _resource_class = 'array'
    array = properties.Array(
        doc='Data, unique values at every point in the mesh',
        shape=('*',),
        dtype=(float, int),
        serializer=array_serializer,
        deserializer=array_download(('*',), (float, int)),
    )

    order = properties.StringChoice(
        doc='Data array order, for data on grid meshes',
        choices={
            'c': ('C-STYLE', 'NUMPY', 'ROW-MAJOR', 'ROW'),
            'f': ('FORTRAN', 'MATLAB', 'COLUMN-MAJOR', 'COLUMN', 'COL')
        },
        default='c',
    )
    colormap = properties.List(
        doc='Colormap applied to data range or categories',
        prop=properties.Color(''),
        min_length=1,
        max_length=256,
        required=False,
        default=properties.undefined,
    )

    def __init__(self, array=None, **kwargs):
        super(DataArray, self).__init__(**kwargs)
        if array is not None:
            self.array = array

    def _nbytes(self, arr=None):
        if arr is None or (isinstance(arr, string_types) and arr == 'array'):
            arr = self.array
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('DataArray cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @properties.observer('array')
    def _reject_large_files(self, change):
        self._validate_file_size(change['name'], change['value'])

    @properties.validator
    def _validate_array(self):
        self._validate_file_size('array', self.array)
        return True

    def _get_dirty_data(self, force=False):
        datadict = super(DataArray, self)._get_dirty_data(force)
        dirty = self._dirty_props
        if 'order' in dirty or force:
            datadict['order'] = self.order
        if self.colormap and ('colormap' in dirty or force):
            datadict['colormap'] = dumps(self.colormap)
        return datadict

    def _get_dirty_files(self, force=False):
        files = super(DataArray, self)._get_dirty_files(force)
        dirty = self._dirty_props
        if 'array' in dirty or force:
            files['array'] = self._props['array'].serialize(self.array)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        data = DataArray(
            title=kwargs['title'],
            description=kwargs['description'],
            order=json['order'],
            array=cls._props['array'].deserialize(
                json['array'],
            )
        )
        if json.get('colormap'):
            data.colormap = json['colormap']
        return data

    @classmethod
    def _build_from_omf(cls, omf_data):
        assert omf_data.__class__.__name__ in ('ScalarData', 'MappedData')
        data = dict(
            location='N' if omf_data.location == 'vertices' else 'CC',
            data=DataArray(
                title=omf_data.name,
                description=omf_data.description,
                array=omf_data.array.array
            )
        )
        return data


class DataCategory(DataArray):
    _resource_class = 'category'  # ...? or same on the backend...?
    array = properties.Array(
        doc='Category index values at every point in the mesh',
        shape=('*',),
        dtype=(int,),
        serializer=array_serializer,
        deserializer=array_download(('*',), (int,)),
    )
    categories = properties.List(
        doc='List of string categories',
        prop=properties.String(''),
        min_length=1,
        max_length=256,
        required=False,
        default=properties.undefined,
    )

    @properties.validator
    def _categories_and_colormap(self):
        if (
                (self.categories and self.colormap) and
                len(self.categories) != len(self.colormap)
        ):
            raise ValueError('categories and colormap must be equal length')

    @properties.validator
    def _categories_and_array(self):
        if min(self.array) < 0:
            raise ValueError('array indices must be >= 0')
        if max(self.array) >= 256:
            raise ValueError('array indices must be < 256')
        if self.categories and max(self.array) >= len(self.categories):
            raise ValueError('array indices must be < len(categories)')
        if self.colormap and max(self.array) >= len(self.colormap):
            raise ValueError('array indices must be < len(colormap)')

    @properties.validator
    def _populate_colormap(self):
        if self.colormap:
            return
        self._categories_and_array()
        self.colormap = self._random_colormap()

    @properties.validator
    def _populate_categories(self):
        if self.categories:
            return
        self._categories_and_array()
        if self.categories:
            cat_len = len(self.categories)
        else:
            cat_len = max(self.array) + 1
        self.categories = ['']*cat_len

    @properties.validator('array')
    def _array_gt_zero(self, change):
        if min(change['value']) < 0:
            raise ValueError('array indices must be >= 0')

    def _get_dirty_data(self, force=False):
        datadict = super(DataCategory, self)._get_dirty_data(force)
        dirty = self._dirty_props
        if 'categories' in dirty or force:
            datadict['categories'] = dumps(self.categories)
        return datadict

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        data = DataCategory(
            title=kwargs['title'],
            description=kwargs['description'],
            order=json['order'],
            array=cls._props['array'].deserialize(
                json['array'],
            ),
            colormap=json['colormap'],
            categories=json['categories'],
        )
        return data

    def _random_colormap(self):
        if self.categories:
            map_len = len(self.categories)
        elif self.array is not None:
            map_len = max(self.array) + 1
        else:
            raise ValueError('categories or array indeces are required for '
                             'random colormap')
        if map_len <= len(properties.basic.COLORS_20):
            return random.sample(properties.basic.COLORS_20, map_len)
        if map_len <= len(properties.basic.COLORS_NAMED):
            return random.sample(list(properties.basic.COLORS_NAMED), map_len)
        if map_len > 256**3:
            raise ValueError('max colormap length must be less than 256**3')
        map_ints = random.sample(range(256**3), map_len)
        def color_from_int(value):
            r = int(value % 256)
            g = int((value-r)/256 % 256)
            b = int(((value-r)/256-g)/256 % 256)
            return (r, g, b)
        return [color_from_int(value) for value in map_ints]


__all__ = ['DataArray', 'DataCategory']



"""line.py contains the Line composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from numpy import max as npmax
from numpy import min as npmin
from numpy import ndarray
from six import string_types
import properties

from .base import BaseMesh
from .base import CompositeResource
from .data import DataArray
from .options import ColorOptions
from .options import Options
from .props import array_serializer, array_download, HasSteno3DProps


class _Mesh1DOptions(Options):
    view_type = properties.StringChoice(
        doc='Display 1D lines or tubes/boreholes/extruded lines',
        choices={
            'line': ('lines', 'thin', '1d'),
            'tube': ('tubes', 'extruded line', 'extruded lines',
                     'borehole', 'boreholes')
        },
        default='line',
        required=False
    )


class _LineOptions(ColorOptions):
    pass


class Mesh1D(BaseMesh):
    """Contains spatial information of a 1D line set"""
    vertices = properties.Array(
        doc='Mesh vertices',
        shape=('*', 3),
        dtype=float,
        serializer=array_serializer,
        deserializer=array_download(('*', 3), (float,)),
    )
    segments = properties.Array(
        doc='Segment endpoint indices',
        shape=('*', 2),
        dtype=int,
        serializer=array_serializer,
        deserializer=array_download(('*', 2), (int,)),
    )
    opts = properties.Instance(
        doc='Options',
        instance_class=_Mesh1DOptions,
        auto_create=True,
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.segments)

    def _nbytes(self, arr=None):
        if arr is None:
            return self._nbytes('segments') + self._nbytes('vertices')
        if isinstance(arr, string_types) and arr in ('segments', 'vertices'):
            arr = getattr(self, arr)
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh1D cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @properties.observer(('segments', 'vertices'))
    def _reject_large_files(self, change):
        self._validate_file_size(change['name'], change['value'])

    @properties.validator
    def _validate_seg(self):
        if npmin(self.segments) < 0:
            raise ValueError('Segments may only have positive integers')
        if npmax(self.segments) >= len(self.vertices):
            raise ValueError('Segments expects more vertices than provided')
        self._validate_file_size('segments', self.segments)
        self._validate_file_size('vertices', self.vertices)
        return True

    def _get_dirty_files(self, force=False):
        files = super(Mesh1D, self)._get_dirty_files(force)
        dirty = self._dirty_props
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self._props['vertices'].serialize(self.vertices)
        if 'segments' in dirty or force:
            files['segments'] = \
                self._props['segments'].serialize(self.segments)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh1D(
            title=kwargs['title'],
            description=kwargs['description'],
            vertices=cls._props['vertices'].deserialize(
                json['vertices'],
            ),
            segments=cls._props['segments'].deserialize(
                json['segments'],
            ),
            opts=json['meta']
        )
        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh1D(
            vertices=(omf_geom.vertices.array +
                      omf_geom.origin +
                      omf_project.origin),
            segments=omf_geom.segments.array
        )
        return mesh


class _LineBinder(HasSteno3DProps):
    """Contains the data on a 1D line set with location information"""
    location = properties.StringChoice(
        doc='Location of the data on mesh',
        choices={
            'CC': ('LINE', 'FACE', 'CELLCENTER', 'EDGE', 'SEGMENT'),
            'N': ('VERTEX', 'NODE', 'ENDPOINT')
        }
    )
    data = properties.Instance(
        doc='Data',
        instance_class=DataArray,
        auto_create=True,
    )


class Line(CompositeResource):
    """Contains all the information about a 1D line set"""
    mesh = properties.Instance(
        doc='Mesh',
        instance_class=Mesh1D,
        auto_create=True,
    )
    data = properties.List(
        doc='Data',
        prop=_LineBinder,
        coerce=True,
        required=False,
    )
    opts = properties.Instance(
        doc='Options',
        instance_class=_LineOptions,
        auto_create=True,
    )

    def _nbytes(self):
        return self.mesh._nbytes() + sum(d.data._nbytes() for d in self.data)

    @properties.validator
    def _validate_data(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'line.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return True


__all__ = ['Line', 'Mesh1D']



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple
from io import BytesIO
from tempfile import NamedTemporaryFile

import numpy as np
from requests import get
import properties


class HasSteno3DProps(properties.HasProperties):

    def __init__(self, **metadata):
        self._dirty_props = set()
        super(HasSteno3DProps, self).__init__(**metadata)

    def _get(self, name):
        value = super(HasSteno3DProps, self)._get(name)
        # Returning a copy of the list maintains backward compatibility
        # until properties has better handling of lists.
        if isinstance(value, list):
            value = [val for val in value]
        return value

    @properties.observer(properties.everything)
    def _mark_dirty(self, change):
        self._dirty_props.add(change['name'])

    def _mark_clean(self, recurse=True):
        self._dirty_props = set()
        if not recurse or getattr(self, '_inside_clean', False):
            return
        self._inside_clean = True
        try:
            props = self._dirty
            for prop in props:
                value = getattr(self, prop)
                if isinstance(value, properties.HasProperties):
                    value._mark_clean()
                if isinstance(value, (list, tuple)):
                    for v in value:
                        if isinstance(v, properties.HasProperties):
                            v._mark_clean()
        finally:
            self._inside_clean = False

    @property
    def _dirty(self):
        if getattr(self, '_inside_dirty', False):
            return set()
        dirty_instances = set()
        self._inside_dirty = True
        try:
            props = self._non_deprecated_props()
            for prop in props:
                value = getattr(self, prop)
                if (isinstance(value, properties.HasProperties) and
                        len(value._dirty) > 0):
                    dirty_instances.add(prop)
                if isinstance(value, (list, tuple)):
                    for v in value:
                        if (isinstance(v, properties.HasProperties) and
                                len(v._dirty) > 0):
                            dirty_instances.add(prop)
        finally:
            self._inside_dirty = False
        return self._dirty_props.union(dirty_instances)

    def _non_deprecated_props(self):
        return {k: v for k, v in self._props.items()
                if not isinstance(v, properties.Renamed)}


def image_download(url, **kwargs):
    im_resp = get(url, timeout=(10, 60))
    if im_resp.status_code != 200:
        raise IOError('Failed to download image.')
    output = BytesIO()
    output.name = 'texture.png'
    for chunk in im_resp:
        output.write(chunk)
    output.seek(0)
    return output


FileProp = namedtuple('FileProp', ['file', 'dtype'])


def array_serializer(data, **kwargs):
    """Convert the array data to a serialized binary format"""
    if isinstance(data.flatten()[0], np.floating):
        use_dtype = '<f4'
        nan_mask = ~np.isnan(data)
        assert np.allclose(
                data.astype(use_dtype)[nan_mask], data[nan_mask]), \
            'Converting the type should not screw things up.'
    elif isinstance(data.flatten()[0], np.integer):
        use_dtype = '<i4'
        assert (data.astype(use_dtype) == data).all(), \
            'Converting the type should not screw things up.'
    else:
        raise TypeError('Must be a float or an int: {}'.format(data.dtype))

    data_file = NamedTemporaryFile('rb+', suffix='.dat')
    data.astype(use_dtype).tofile(data_file.file)
    data_file.seek(0)
    return FileProp(data_file, use_dtype)


class array_download(object):

    def __init__(self, shape, dtype):
        self.shape = shape
        self.dtype = dtype

    def __call__(self, url, **kwargs):
        arr_resp = get(url, timeout=(10, 60))
        if arr_resp.status_code != 200:
            raise IOError('Failed to download array.')
        data_file = NamedTemporaryFile()
        for chunk in arr_resp:
            data_file.write(chunk)
        data_file.seek(0)
        if self.dtype[0] is int:
            dtype = '<i4'
        elif self.dtype[0] is float:
            dtype = '<f4'
        else:
            raise ValueError('Invalid dtype {}'.format(self.dtype))
        arr = np.fromfile(data_file.file, dtype)
        unknown_dim = len(arr)
        for dim in self.shape:
            if dim == '*':
                continue
            unknown_dim /= dim
        if '*' in self.shape:
            assert abs(unknown_dim - int(unknown_dim)) < 1e-9, 'bad shape'
            shape = tuple(
                (int(unknown_dim) if dim == '*' else dim for dim in self.shape)
            )
        else:
            assert abs(unknown_dim) < 1e-9, 'bad shape'
            shape = self.shape
        arr = arr.reshape(shape)
        data_file.close()
        return arr



"""volume.py contains the Volume composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from numpy import ndarray
from six import string_types
import properties

from .base import BaseMesh
from .base import CompositeResource
from .data import DataArray
from .options import ColorOptions
from .options import MeshOptions
from .props import HasSteno3DProps


class _Mesh3DOptions(MeshOptions):
    pass


class _VolumeOptions(ColorOptions):
    pass


class Mesh3DGrid(BaseMesh):
    """Contains spatial information of a 3D grid volume."""

    h1 = properties.Array(
        doc='Tensor cell widths, x-direction',
        shape=('*',),
        dtype=(float, int)
    )
    h2 = properties.Array(
        doc='Tensor cell widths, y-direction',
        shape=('*',),
        dtype=(float, int)
    )
    h3 = properties.Array(
        doc='Tensor cell widths, z-direction',
        shape=('*',),
        dtype=(float, int)
    )
    x0 = properties.Renamed('O')
    O = properties.Vector3(
        doc='Origin vector',
        default=[0., 0., 0.]
    )
    U = properties.Vector3(
        doc='Orientation of h1 axis',
        default='X'
    )
    V = properties.Vector3(
        doc='Orientation of h2 axis',
        default='Y'
    )
    W = properties.Vector3(
        doc='Orientation of h3 axis',
        default='Z'
    )
    opts = properties.Instance(
        doc='Mesh3D Options',
        instance_class=_Mesh3DOptions,
        auto_create=True,
    )

    @property
    def nN(self):
        """ get number of nodes """
        return (len(self.h1)+1) * (len(self.h2)+1) * (len(self.h3)+1)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.h1) * len(self.h2) * len(self.h3)

    def _nbytes(self, arr=None):
        filenames = ('h1', 'h2', 'h3', 'O', 'U', 'V', 'W')
        if arr is None:
            return sum(self._nbytes(fn) for fn in filenames)
        if isinstance(arr, string_types) and arr in filenames:
            if getattr(self, arr, None) is None:
                return 0
            arr = getattr(self, arr)
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh3DGrid cannot calculate the number of '
                         'bytes of {}'.format(arr))

    def _get_dirty_data(self, force=False):
        datadict = super(Mesh3DGrid, self)._get_dirty_data(force)
        dirty = self._dirty_props
        if force or ('h1' in dirty or 'h2' in dirty or 'h3' in dirty):
            datadict['tensors'] = dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
                h3=self.h3.tolist()
            ))
        if force or any([item in dirty for item in
                         ['O', 'U', 'V', 'W', 'h1', 'h2', 'h3']]):
            datadict['OUVZ'] = dumps(dict(
                O=self.O.tolist(),
                U=self.U.as_length(self.h1.sum()).tolist(),
                V=self.V.as_length(self.h2.sum()).tolist(),
                Z=self.W.as_length(self.h3.sum()).tolist()
            ))
        return datadict

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh3DGrid(
            title=kwargs['title'],
            description=kwargs['description'],
            h1=json['tensors']['h1'],
            h2=json['tensors']['h2'],
            h3=json['tensors']['h3'],
            O=json['OUVZ']['O'],
            U=json['OUVZ']['U'],
            V=json['OUVZ']['V'],
            W=json['OUVZ']['Z'],
            opts=json['meta']
        )
        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh3DGrid(
            h1=omf_geom.tensor_u,
            h2=omf_geom.tensor_v,
            h3=omf_geom.tensor_w,
            O=omf_geom.origin + omf_project.origin,
            U=omf_geom.axis_u,
            V=omf_geom.axis_v,
            W=omf_geom.axis_w
        )
        return mesh


class _VolumeBinder(HasSteno3DProps):
    """Contains the data on a 3D volume with location information"""
    location = properties.StringChoice(
        doc='Location of the data on mesh',
        choices={
            'CC': ('CELLCENTER'),
            # 'N': ('NODE', 'VERTEX', 'CORNER')
        },
        default='CC'
    )
    data = properties.Instance(
        doc='Data',
        instance_class=DataArray,
        auto_create=True,
    )


class Volume(CompositeResource):
    """Contains all the information about a 3D volume"""
    mesh = properties.Instance(
        doc='Mesh',
        instance_class=Mesh3DGrid,
        auto_create=True,
    )
    data = properties.List(
        doc='Data',
        prop=_VolumeBinder,
        coerce=True,
        required=False
    )
    opts = properties.Instance(
        doc='Options',
        instance_class=_VolumeOptions,
        auto_create=True,
    )

    def _nbytes(self):
        return self.mesh._nbytes() + sum(d.data._nbytes() for d in self.data)

    @properties.validator
    def _validate_data(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location == 'CC'  # in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'volume.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return True


__all__ = ['Volume', 'Mesh3DGrid']



#!/usr/bin/env python
'''
    steno3d: Client library for Steno3D & steno3d.com
'''

from distutils.core import setup
from setuptools import find_packages

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Physics',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS',
    'Natural Language :: English',
]

with open('README.rst') as f:
    LONG_DESCRIPTION = ''.join(f.readlines())

setup(
    name='steno3d',
    version='0.3.2b0',
    packages=find_packages(),
    install_requires=[
        'numpy>=1.7',
        'pypng',
        'requests',
        'six',
        'properties>=0.3.3',
        'vectormath',
    ],
    author='ARANZ Geo Limited',
    author_email='support@steno3d.com',
    description='steno3d',
    long_description=LONG_DESCRIPTION,
    keywords='visualization',
    url='https://steno3d.com/',
    download_url='http://github.com/3ptscience/steno3dpy',
    classifiers=CLASSIFIERS,
    platforms=['Windows', 'Linux', 'Solaris', 'Mac OS-X', 'Unix'],
    license='MIT License',
    use_2to3=False,
)



"""options.py defines the base option classes for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
import properties

from .props import HasSteno3DProps


class Options(HasSteno3DProps):
    """Generic options for all steno3d resources"""

    @property
    def _json(self):
        """returns json representation of options"""
        opts_json = {}
        for key in self._props:
            opts_json[key] = getattr(self, key)
        return dumps(opts_json)

    @properties.validator(properties.everything)
    def _opt_defaulter(self, change):
        if change['value'] is None:
            change['value'] = self._props[change['name']].default


class ColorOptions(Options):
    """Options related to resource display color"""
    color = properties.Color(
        doc='Solid color',
        default='random',
        required=False
    )
    opacity = properties.Float(
        doc='Opacity',
        default=1.,
        min=0.,
        max=1.,
        required=False
    )


class MeshOptions(Options):
    """Options related to mesh display"""
    wireframe = properties.Bool(
        doc='Wireframe',
        default=False,
        required=False
    )
